use mastodon::authorize_fedi;
use rand::seq::SliceRandom;
use std::convert::TryInto;
use std::fs::File;
use std::io::{BufWriter, Cursor, Write};
use tracing::{event, Level};
mod mastodon;
mod svg;
mod twitter;
mod wiki;
use clap::{Parser, Subcommand};

use std::borrow::Cow;
use twitter::*;
use wiki::*;

use crate::mastodon::toot;

static APP_USER_AGENT: &str = concat!(
    "bot_",
    env!("CARGO_PKG_NAME"),
    "/",
    env!("CARGO_PKG_VERSION"),
    " ",
    "reqwest",
);

type StdError<T> = Result<T, Box<dyn std::error::Error>>;

#[derive(Parser)]
#[clap(author, version, about)]
struct Cli {
    #[clap(subcommand)]
    command: Commands,
}

#[derive(Subcommand)]
enum Commands {
    /// Authorize the twitter application to acccess a user's account
    AuthorizeTwitter,
    /// Scrape images from the category on wikimedia commons
    ScrapeCategory,
    /// Scrape images from the iso7010 wikipedia page
    ScrapeWeb,
    /// List tweets from the authed user's timeline
    ListTweets,
    /// Authorize against a pleroma server
    AuthorizeFedi,
    /// Run the bot - scrape, pick a random entry and toot it
    RunBot {
        #[clap(short, long, action)]
        dry_run: bool,
        #[clap(short, long, value_parser)]
        target: Option<String>,
    },
    /// Print details about the currently authed user
    Whoami,
    /// Download and dump all images to disk
    Dump,
}

#[tokio::main]
async fn main() -> StdError<()> {
    tracing_subscriber::fmt::init();
    let cli = Cli::parse();
    match &cli.command {
        Commands::AuthorizeTwitter => do_authorize().await,
        Commands::ScrapeCategory => do_scrape_category().await,
        Commands::ScrapeWeb => do_scrape_web().await,
        Commands::ListTweets => do_list_tweets().await,
        Commands::Whoami => do_whoami().await,
        Commands::AuthorizeFedi => authorize_fedi().await,
        Commands::RunBot { dry_run, target } => run_bot(*dry_run, target.to_owned()).await,
        Commands::Dump => do_dump().await,
    }
}

async fn do_whoami() -> StdError<()> {
    let user_token = user_token_from_env();

    let user: serde_json::Value = twitter_api(
        TwitterEndpoint::VerifyCredentials.try_into()?,
        Some(&user_token),
        APIAction::Get,
        &[],
    )
    .await?
    .json()
    .await?;
    println!("User @{}, (id: {})", user["screen_name"], user["id"]);
    Ok(())
}

async fn do_list_tweets() -> StdError<()> {
    let user_token = user_token_from_env();

    let user = twitter_api(
        TwitterEndpoint::VerifyCredentials.try_into()?,
        Some(&user_token),
        APIAction::Get,
        &[],
    )
    .await?
    .json::<serde_json::Value>()
    .await?;

    let id = user["id"].as_u64().unwrap().to_string();
    let mut timeline = vec![];
    let mut last_id: Option<u64> = None;
    loop {
        event!(Level::INFO, last_id, "Fetching chunk of tweets...");
        let mut params: Vec<(&str, Cow<str>)> = vec![
            ("count", "200".into()),
            ("exclude_replies", "true".into()),
            ("include_rts", "false".into()),
            ("trim_user", "true".into()),
            ("user_id", Cow::Borrowed(&id)),
        ];
        if let Some(since_id) = last_id {
            // Get next chunk of tweets before those we've fetched
            params.push(("max_id", (since_id - 1).to_string().into()))
        }
        let timeline_chunk = twitter_api(
            reqwest::Url::parse_with_params(&TwitterEndpoint::UserTimeline.to_string(), params)?,
            Some(&user_token),
            APIAction::Get,
            &[],
        )
        .await?
        .json::<serde_json::Value>()
        .await?;
        let chunk = timeline_chunk.as_array().unwrap().to_owned();
        event!(Level::INFO, count = chunk.len(), "Got tweets.");
        if chunk.is_empty() {
            break;
        }
        last_id = Some(
            chunk.last().unwrap().as_object().unwrap()["id"]
                .as_u64()
                .unwrap(),
        );
        timeline.extend(chunk);
    }
    for tweet in timeline {
        let tweet = tweet.as_object().unwrap();
        println!("{}, \"{}\"", tweet["id"], tweet["text"]);
    }
    Ok(())
}

async fn do_scrape_category() -> StdError<()> {
    let mut files =
        get_files_in_category("Category:ISO_7010_safety_signs_(vector_drawings)").await?;
    files.sort();
    for f in files {
        println!("{}", f);
    }

    Ok(())
}

async fn do_scrape_web() -> StdError<()> {
    let mut files: Vec<_> = scrape_web()
        .await?
        .into_iter()
        .map(|(_, file)| file)
        .collect();
    files.sort();
    for f in files {
        println!("{}", f);
    }

    Ok(())
}

fn get_client(headers: Option<reqwest::header::HeaderMap>) -> StdError<reqwest::Client> {
    let mut c = reqwest::Client::builder().user_agent(APP_USER_AGENT);
    if let Some(headers) = headers {
        c = c.default_headers(headers);
    }
    Ok(c.build()?)
}

async fn do_dump() -> StdError<()> {
    let all = scrape_web().await?;
    let filenames: Vec<_> = all.iter().map(|(_, filename)| filename.as_str()).collect();
    let metas = get_file_metadata(&filenames).await?;
    // for m in &metas {
    //     let bytes = client.get(m.url.to_string()).await?.bytes().await?;
    // }
    let file = File::create("dump.json")?;
    let mut writer = BufWriter::new(file);
    serde_json::to_writer_pretty(&mut writer, &metas)?;
    writer.flush()?;
    let count = metas.len();
    event!(Level::INFO, count, "Dumped records to json file");
    Ok(())
}

async fn run_bot(dry_run: bool, target: Option<String>) -> StdError<()> {
    let all = scrape_web().await?;
    let (title, filename) = if let Some(target) = target {
        all.iter()
            .find(|(title, _)| title.to_lowercase().contains(&target.to_lowercase()))
    } else {
        all.choose(&mut rand::thread_rng())
    }
    .ok_or("got no images m8")?;
    event!(Level::INFO, title, filename, "Picked random thing");
    let client = get_client(None)?;
    event!(Level::INFO, "Fetching metadata...");
    // TODO: could crash, probably doesn't matter
    let meta = get_file_metadata(&[filename.as_str()]).await?.remove(0);
    event!(Level::INFO, %meta, "Got metadata");
    event!(Level::INFO, url = meta.url.to_string(), "Fetching image");
    let svg = client.get(meta.url).send().await?.bytes().await?;

    let text = format!(
        "{}\n\nImage source: {}\nAuthor: Wikimedia Commons user {}\n{}{}",
        title,
        meta.html_url,
        meta.author,
        meta.license_short_name,
        meta.license_url
            .map_or("".to_owned(), |u| format!(" ({})", u))
    );

    if !dry_run {
        // Render the image nice and big for twitter
        let img = svg::render_svg(&svg, 1000, false)?;
        let mut buf = Cursor::new(Vec::new());
        img.write_to(&mut buf, image::ImageFormat::Png)?;
        toot(&text, Some(Box::new(buf))).await?;
    } else {
        // Render the image smaller for output to terminal
        let img = svg::render_svg(&svg, 128, true)?;
        println!("Dry run - would tweet:\n \"{}\"", text);
        viuer::print(
            &img,
            &viuer::Config {
                absolute_offset: false,
                width: Some(32),
                ..Default::default()
            },
        )?;
    }

    Ok(())
}
